
/*
** mszp reads the fileformat used by InfoView in MSVC
*/

#include <stdio.h>
#include <stdlib.h>
#include <sys/types.h>
#include <sys/stat.h>
#include "support.h"
#include "ivt.h"

int read_a_field(FILE *mszp,int *n)
{
	int ch, val=0;
	*n=0;

	ch = fgetc(mszp);
	ch&=0x7f;
	val += (ch<<(7*(*n)));
	ch = fgetc(mszp);
	(*n)++;

	while(ch&0x80)
	{
		ch&=0x7f;
		val += (ch<<(7*(*n)));
		ch = fgetc(mszp);
		(*n)++;
	}

	ch&=0x7f;
	val += (ch<<(7*(*n)));

	return val;
}

int read_b_field(FILE *mszp,int n)
{
	int ch, val=0;
	n=0;

	ch = fgetc(mszp);
	ch&=0x7f;
	val += (ch<<(7*n));
	ch = fgetc(mszp);
	n++;

	while(ch)
	{
		ch&=0x7f;
		val += (ch<<(7*n));
		ch = fgetc(mszp);
		n++;
	}

	ch&=0x7f;
	val += (ch<<(7*n));

	return val;
}

/* read a field following the filename */
int read_field(FILE *mszp,int *n) 
	{
    int finished=0, ch, val=0;
	*n=0;

    while(!feof(mszp)) {
        ch = fgetc(mszp);
        /* printf("%02X ",ch); */
        if(ch==0)
            return -1;
        if(!(ch&0x80))
            finished++;
        ch&=0x7f;
        val += (ch<<(7*(*n)));
        (*n)++;
        if(*n>5)
            return -1;
        if(finished) {
            /* printf("(%X) ",val); */
            return val;
        }
    }
return -1;
}

/* create a linked list of windows zip directory entries */
void add_dirent(mszpentry **chain,mszpentry *entry) {
    mszpentry *e,*t;

    e = (mszpentry *) malloc(sizeof(mszpentry));
    memcpy(e,entry,sizeof(mszpentry));
    e->next=NULL;
    if(*chain) {
        for(t=*chain; t->next; t=t->next);
        t->next=e;
    }
    else {
        *chain=e;
    }
}

void wvGetmszpentryChain(FILE *ivt,mszpentry **chain,globalivtidx *index)
{
	int i;
	unsigned int j=0;
	ivtidx idx;
	mszpentry entry;
	while(j<index->fullentries)
	{
		if (wvGetivtidx(ivt,&idx))
			return;
		for(i=0; i<idx.entries; i++,j++) 
			{
			if(!wvGetmszpentry(ivt,&entry)) 
				{
				entry.counter=i;
				add_dirent(chain,&entry);
				}
			}
		fseek(ivt,idx.nextoffset,SEEK_CUR);
	}
}

void wvReleasemszpentryChain(mszpentry **chain) 
{
    mszpentry *e,*t;
	e = *chain;

	while (e) 
	{
		t = e->next;
		free(e);
		e = t;
	}
}

/* read a complete entry name, offset, length from the contents section */
int wvGetmszpentry(FILE *mszp,mszpentry *entry) {
    int len=fgetc(mszp);
    int ch,i,n;

    if(len==EOF)
        return -1;
    if(!len)
        return -1;

    for(i=0; i<len; i++) {
        if((ch=fgetc(mszp))==EOF)
            break;
        if(ch&0x80) {
            fseek(mszp,-1,SEEK_CUR);
            len=i-1;
            break;
        }
        entry->name[i]=ch;
    }
    entry->name[len]=0;

    entry->offset = read_a_field(mszp,&n);
    if(entry->offset == -1)
        return -1;
    entry->length = read_b_field(mszp,n);
    if(entry->length == -1)
        return -1;

    return 0;
}

void print_entry(mszpentry *entry) {
    printf("@%06X %06X %s\n",
           entry->offset,
           entry->length,
           entry->name);
}



int print_header(unsigned char *h) {
    int i,length;

    for(i=0; i<0x0e; i+=2)
       printf("%02X%02X ",h[i+1],h[i]);
    length = h[0xa] + h[0xb]*0x100;
    printf(" %X",length*2);
    return length;
}
#if 0
void print_content(int length, FILE *mszp) {
    int i;
    unsigned char buffer[0x10];

    /* dump the gif! */
    fread(buffer, sizeof buffer,1, mszp);
#ifdef DUMP_GIF
    if(buffer[5]=='G' && buffer[6]=='I' && buffer[7]=='F') {
        FILE *f = fopen("/home/mike/ext.gif","w+");
        printf("Dumping GIF...");
        fwrite(&buffer[5],sizeof buffer-5,1,f);
        length-=5;
        while(length>0) {
            int l=fread(buffer,1,sizeof buffer,mszp);
            fwrite(buffer,1,l,f);
            length-=l;
        }
        fclose(f);
        exit(0);
    }
#endif
    for(i=0; i<sizeof buffer; i++)
        printf("%02X ",buffer[i]);
    for(i=0; i<sizeof buffer; i++)
        printf("%c", isprint(buffer[i])?buffer[i]:'.');
    printf("\n");
}
#endif

int check_name(char *name) {
    char *p;

    /* skip any leading slashes - create in current directory */
    while (*name == '/')
        name++;

    p=name;

    while(*p) {
        if(*p=='/')
            p++;

        /* find end */
        while(*p && (*p!='/'))
            p++;

        /* if this is a directory, create it */
        if(*p == '/') {
            *p = 0;
            mkdir(name,0x1FF); /* fails if the directory already exists */
            *p = '/';
        }
    }

    return 0;
}

int html_name(char *name) 
{
	/*fprintf(stderr,"%s\n",name+strlen(name)-4);*/
	if (strcasecmp(name+strlen(name)-4,".htm") == 0)
		return(check_name(name));
	else if (strcasecmp(name+strlen(name)-4,"html") == 0)
		return(check_name(name));
	else if (strcasecmp(name+strlen(name)-4,".css") == 0)
		return(check_name(name));
	else if (strcasecmp(name+strlen(name)-4,".png") == 0)
		return(check_name(name));
	else if (strcasecmp(name+strlen(name)-4,".jpg") == 0)
		return(check_name(name));
	else if (strcasecmp(name+strlen(name)-4,".gif") == 0)
		return(check_name(name));
	return(-1);
}



/* copy from the archive into a directory file */
int extract_file(FILE *mszp, int length, char *name) {
    FILE *out;
    int len,written=0;
    static char buf[0x100];

    while(*name && (name[0]=='/'))
        name++;
    out = fopen(name,"wb");
    if(!out)
        return -1;

    for(written=0; written<length; ) 
	{
        len = fread(buf,1,MIN((int)sizeof buf,length-written),mszp);
        if(!fwrite(buf,1,len,out))
            break;
        written+=len;
    }

    fclose(out);

    /*printf("Wrote %d bytes to %s\n",written,name);*/

    return 0;
}

/* copy from the archive into a directory file */
int extract_cfile(FILE *ivt, int length, char *name) 
{
    FILE *out;
    int len,written=0,err=-1;
	U8 uncompr[99999];
    static char buf[0x100];
	MSZP mszp;
	
	wvInitMSZP(&mszp);

    while(*name && (name[0]=='/'))
        name++;
    out = tmpfile();
    if(!out)
        return -1;

    for(written=0; written<length; ) {
        len = fread(buf,1,MIN((int)sizeof buf,length-written),ivt);
        if(!fwrite(buf,1,len,out))
            break;
        written+=len;
    }
	rewind(out);
	wvGetMSZP(&mszp,out);
	fclose(out);

	if (!mszp.compr)
	{
		fprintf(stderr,"No data to decompress\n");
		return(-1);
	}

	err = wvInflatemszp(mszp.compr,uncompr);

	if (err != 0)
	{
		fprintf(stderr, "error: %d\n", err);
		return(-1);
	}

    out = fopen(name,"wb");
	fwrite(uncompr, mszp.uncomprLen, 1, out);
    fclose(out);

    /*printf("Wrote %d bytes to %s\n",written,name);*/

    return 0;
}

/* 
** dump a file's contents to the harddisk 
*/
int wvDumpFile(FILE *mszp,mszpentry *t) 
{
    /*
    ** read the magic number, so we can decide what type of file it is
    */
    fseek(mszp,t->offset,SEEK_SET);

    if(!check_name(t->name))
        return extract_file(mszp,t->length,t->name);

    return -1;
}

int wvOutputHtml(FILE *mszp,mszpentry *t) 
{
    /*
    ** read the magic number, so we can decide what type of file it is
    */
    fseek(mszp,t->offset,SEEK_SET);

    if(!html_name(t->name))
    	 {
			printf("-->%s\n",t->name);
			return extract_cfile(mszp,t->length,t->name);
		 }
	else
		printf("ignoring \"%s\"\n",t->name);

    return -1;
}

void wvGetivthdr(FILE *fd,ivthdr *item)
{
   item->Magic = read_32ubit(fd);
   item->DirectoryStart = read_32ubit(fd);
   item->uk1 = read_32ubit(fd);
   item->uk2 = read_32ubit(fd);
   item->uk3 = read_32ubit(fd);
   item->filesz = read_32ubit(fd);
   item->uk4 = read_32ubit(fd);
   item->uk5 = read_32ubit(fd);
   item->uk6 = read_32ubit(fd);
   item->uk7 = read_32ubit(fd);
}

void wvGetglobalivtidx(FILE *fd,globalivtidx *item)
{
	int i;
	item->magic = read_16ubit(fd);
	item->flags = read_16ubit(fd);
	for(i=0;i<10;i++)
		item->uk1[i] = read_32ubit(fd);
	item->fullentries = read_32ubit(fd);
}

int wvGetivtidx(FILE *fd,ivtidx *item)
{
	item->nextoffset = read_16ubit(fd);
	item->entries = read_16ubit(fd);
	if (item->entries > 0x3300)
	{
		fprintf(stderr,"Halting, probable last record\n");
		return(1);
	}
	item->uk2[0] = read_32ubit(fd);
	if (item->entries < 4) /*reaching for straws, see comment A below*/
		return(0);
	item->uk2[1] = read_32ubit(fd);
	return(0);
}

/*          00 00 [nof] [len] [...
 * 
 * 00ac168  00 00 fd 0f a1 00 00 00  00ac168 ( . . . . . . .
 * 00ac170  00 00 03 00 00 00 14 2f  00ac170 . . . . . . . /
 * 			................] begin
 *
 *
 * 00ae168  00 00 ae 1f 03 00 00 00  00ae168 . . . . . . . .
 * 00ae170  00 00 14 2f 72 74 66 2f  00ae170 . . . / r t f /
 *
 * 00b0168  00 00 0b 10 9f 00 01 00  00b0168 . . . . . . . .
 * 00b0170  00 00 04 00 00 00 14 2f  00b0170 . . . . . . . /
 *
 * 00b2168  00 00 9b 05 0e 01 03 00  00b2168 . . . . . . . .
 * 00b2170  00 00 ff ff ff ff 13 2f  00b2170 . . . . . . . /
 *
 *
 * 00b4168  00 00 ad 5f 6b 34 00 00  00b4168 . . . _ k 4 . .
 * 00b4170  00 00 00 00 00 00 00 00  00b4170 . . . . . . . .
 *
 */
